/****************************************************************************
 * Include Files
 */
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>

#include "FreeRTOS.h"
#include "event_groups.h"
#include "semphr.h"
#include "task.h"

#include "app_error.h"
#include "app_util.h"
#include "app_util_platform.h"
#include "ble_data.h"
#include "ble_srv_common.h"
#include "gatt_service_defs.h"
#include "crc16.h"
#include "nordic_common.h"
#include "nrf_ble_gatt.h"
#include "nrf_delay.h"
#include "nrf_log.h"

/****************************************************************************
 * Constants
 */

/* Number of bytes of overhead for a notification. The largest possible
 * notification payload is MTU - ATT_NOTIFY_OVERHEAD.
 */
#define ATT_NOTIFY_OVERHEAD (3)

#define EVENT_DISCONNECTED (1<<0)
#define EVENT_BLE_READY (1<<1)

#define WAIT_FOR_BLE_READY_TIMEOUT (pdMS_TO_TICKS(500))

/****************************************************************************
 * Local Types
 */

typedef PACKED_STRUCT {
    uint16_t sequence;
    uint16_t crc;
} NOTIFY_HEADER_T;

typedef PACKED_STRUCT {
    NOTIFY_HEADER_T header;
    uint8_t payload[NRF_SDH_BLE_GATT_MAX_MTU_SIZE -
        ATT_NOTIFY_OVERHEAD - sizeof(NOTIFY_HEADER_T)];
} NOTIFY_T;

/****************************************************************************
 * Local Variables
 */

static uint16_t att_mtu;

static uint16_t block_crc;

/* Indicates that we are currently sending a block of data. */
static SemaphoreHandle_t busy_mutex;

static uint16_t conn_handle;

static ble_gatts_char_handles_t data_char_handles;

static EventGroupHandle_t event_group;

static SemaphoreHandle_t module_mutex;

static NOTIFY_T notify;

static size_t notify_payload_len;

static uint16_t sequence;

/****************************************************************************
 * Local Functions
 */
static uint32_t add_char_data(uint8_t base_uuid_type, uint16_t service_handle);
static bool is_cccd_configured(void);
static void module_mutex_give(void);
static void module_mutex_take(void);
static size_t notify_payload_free(void);
static size_t notify_payload_size(void);
static void on_connect(ble_evt_t const *p_ble_evt);
static void on_disconnect(ble_evt_t const *p_ble_evt);
static void on_tx_complete(ble_evt_t const *p_ble_evt);
static bool send_notify(bool final_notify);
static bool wait_for_ble_ready(void);

static uint32_t
add_char_data(uint8_t base_uuid_type, uint16_t service_handle)
{
    uint32_t err_code;

    ble_add_char_params_t params = {0};
    params.uuid = TEST_DATA_UUID;
    params.uuid_type = base_uuid_type;
    params.max_len = NRF_SDH_BLE_GATT_MAX_MTU_SIZE;
    params.is_var_len = true;
    params.char_props.notify = 1;
    params.cccd_write_access = SEC_OPEN;

    err_code = characteristic_add(service_handle, &params, &data_char_handles);
    return err_code;
}

void
ble_data_finish()
{
    ble_data_send(NULL, 0, true);

    xSemaphoreGive(busy_mutex);
}

uint32_t
ble_data_init(uint8_t base_uuid_type, uint16_t service_handle)
{
    uint32_t err_code;

    err_code = add_char_data(base_uuid_type, service_handle);
    if(err_code != NRF_SUCCESS) {
        return err_code;
    }

    module_mutex = xSemaphoreCreateMutex();
    if (NULL == module_mutex) {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }

    busy_mutex = xSemaphoreCreateMutex();
    if(NULL == busy_mutex) {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }

    event_group = xEventGroupCreate();
    if(NULL == event_group) {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }

    APP_ERROR_CHECK_BOOL(event_group != NULL);

    xEventGroupSetBits(event_group, EVENT_DISCONNECTED);

    return NRF_SUCCESS;
}

void
ble_data_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context)
{
    module_mutex_take();

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_ble_evt);
            break;
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_ble_evt);
            break;
        case BLE_GATTS_EVT_HVN_TX_COMPLETE:
            on_tx_complete(p_ble_evt);
            break;
        default:
            break;
    }

    module_mutex_give();
}

void
ble_data_on_ble_gatt_evt(nrf_ble_gatt_t *p_gatt,
        nrf_ble_gatt_evt_t const *p_evt)
{
    switch (p_evt->evt_id) {
        case NRF_BLE_GATT_EVT_ATT_MTU_UPDATED:
            att_mtu = p_evt->params.att_mtu_effective;
            break;
        default:
            break;
    }
}

bool
ble_data_send(const uint8_t *data, size_t len, bool final_data)
{
    module_mutex_take();

    if (conn_handle == BLE_CONN_HANDLE_INVALID) {
        module_mutex_give();
        return false;
    }

    do {
        size_t to_copy = MIN(len, notify_payload_free());
        if (data != NULL) {
            memcpy(&notify.payload[notify_payload_len], data, to_copy);
            data = &data[to_copy];
        }
        notify_payload_len += to_copy;
        len -= to_copy;

        if (final_data || notify_payload_free() == 0) {
            module_mutex_give();
            if (!wait_for_ble_ready()) {
                return false;

            }
            module_mutex_take();
            if (!send_notify(final_data && len == 0)) {
                continue;
            }
            if (final_data && len == 0) {
                break;
            }
        }
    } while (len > 0 || final_data);

    module_mutex_give();
    return true;
}

bool
ble_data_start()
{
    bool can_start;

    module_mutex_take();

    if (conn_handle == BLE_CONN_HANDLE_INVALID) {
        can_start = false;
    } else if (!is_cccd_configured()) {
        can_start = false;
    } else {
        BaseType_t ret = xSemaphoreTake(busy_mutex, 0);
        can_start = ret == pdTRUE;
    }

    if (can_start) {
        block_crc = 0xffff;
        sequence = 0;
        notify_payload_len = 0;
    }

    module_mutex_give();

    return can_start;
}

static bool
is_cccd_configured(void)
{
    uint32_t err_code;
    uint8_t  cccd_value_buf[BLE_CCCD_VALUE_LEN];
    bool is_notify_enabled = false;

    ble_gatts_value_t value = {
        .len = BLE_CCCD_VALUE_LEN,
        .offset = 0,
        .p_value = cccd_value_buf
    };
    err_code = sd_ble_gatts_value_get(conn_handle,
            data_char_handles.cccd_handle, &value);

    if (err_code != NRF_SUCCESS) {
        return false;
    }

    is_notify_enabled = ble_srv_is_notification_enabled(cccd_value_buf);

    return is_notify_enabled;
}

static void
module_mutex_give(void)
{
    BaseType_t ret;
    ret = xSemaphoreGive(module_mutex);
    APP_ERROR_CHECK_BOOL(ret == pdTRUE);
}

static void
module_mutex_take(void)
{
    BaseType_t ret;
    ret = xSemaphoreTake(module_mutex, portMAX_DELAY);
    APP_ERROR_CHECK_BOOL(ret == pdTRUE);
}

static size_t
notify_payload_free(void)
{
    return notify_payload_size() - notify_payload_len;
}

/**
 * Maximum number of payload bytes that can be sent in a notification.
 *
 * This depends on the ATT MTU that is in effect for the current connection.
 */
static size_t
notify_payload_size(void)
{
    return att_mtu - ATT_NOTIFY_OVERHEAD - sizeof(NOTIFY_HEADER_T);
}

static void
on_connect(ble_evt_t const *p_ble_evt)
{
    conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
    att_mtu = BLE_GATT_ATT_MTU_DEFAULT;
    xEventGroupClearBits(event_group, EVENT_DISCONNECTED);
    xEventGroupSetBits(event_group, EVENT_BLE_READY);
}

static void
on_disconnect(ble_evt_t const *p_ble_evt)
{
    conn_handle = BLE_CONN_HANDLE_INVALID;
    xEventGroupSetBits(event_group, EVENT_DISCONNECTED);
    xSemaphoreGive(busy_mutex);
}

static void
on_tx_complete(ble_evt_t const *p_ble_evt)
{
    xEventGroupSetBits(event_group, EVENT_BLE_READY);
}

static bool
send_notify(bool final_notify)
{
    uint32_t err_code;

    notify.header.sequence = sequence + (final_notify ? 0x8000 : 0);
    notify.header.crc = crc16_compute(notify.payload, notify_payload_len, &block_crc);

    uint16_t tx_len = notify_payload_len + sizeof(NOTIFY_HEADER_T);
    ble_gatts_hvx_params_t const hvx = {
        .type = BLE_GATT_HVX_NOTIFICATION,
        .handle = data_char_handles.value_handle,
        .p_data = (uint8_t *)&notify,
        .p_len = &tx_len
    };
    err_code = sd_ble_gatts_hvx(conn_handle, &hvx);
    if (err_code == NRF_SUCCESS) {
        notify_payload_len = 0;
        block_crc = notify.header.crc;
        sequence++;
        APP_ERROR_CHECK_BOOL(sequence < 0x8000);
        return true;
    } else if (err_code == NRF_ERROR_RESOURCES) {
        xEventGroupClearBits(event_group, EVENT_BLE_READY);
        return false;
    } else {
        NRF_LOG_ERROR("sd_ble_gatts_hvx() failed: 0x%x", err_code);
        return false;
    }
}

static bool
wait_for_ble_ready()
{
    EventBits_t bits = xEventGroupWaitBits(event_group,
            EVENT_BLE_READY | EVENT_DISCONNECTED, false, false,
            WAIT_FOR_BLE_READY_TIMEOUT);

    bool is_ready = bits & EVENT_BLE_READY;
    xEventGroupSetBits(event_group, EVENT_BLE_READY);
    bool is_connected = !(bits & EVENT_DISCONNECTED);

    return is_ready && is_connected;
}

