'use strict';

const ble = require('pc-ble-driver-js');

const adapterFactory = ble.AdapterFactory.getInstance(undefined, { enablePolling: false });

const BLE_UUID_CCCD = '2902';
const TEST_SERVICE_UUID = 'DE77100090E111E89A5A34F39A69480C';
const TEST_DATA_UUID = 'DE77100390E111E89A5A34F39A69480C';

class DeviceBLE {
  constructor(adapter, device) {
    this.adapter = adapter;
    this.device = device;
    this.connected = false;
  }

  setup() {
    return this.discoverService()
      .then(() => this.discoverCharacteristics())
      .then(() => this.discoverDataCCCD())
      .then(() => this.setDataCCCD())
      .then(() => {
        const onDisconnect = (device) => {
          if (device.instanceId === this.device.instanceId) {
            this.adapter.removeListener('deviceDisconnected', onDisconnect);
            this.onDisconnected();
          }
        };
        this.adapter.on('deviceDisconnected', onDisconnect);
        this.adapter.on('characteristicValueChanged', attribute => this.onCharacteristicValueChanged(attribute));
        this.connected = true;
        return this;
      });
  }

  onCharacteristicValueChanged(attribute) {
    if (attribute.instanceId === this.dataChar.instanceId) {
      this.onDataCharValueChanged(attribute);
    }
  }

  onDataCharValueChanged(attribute) {
    const v = attribute.value;
    const n = v[0] + (v[1] << 8);
    console.log('Notify ' + n);
    if (n & 0x8000) {
        console.log('Got the last notify')
    }
  }

  onDisconnected() {
    this.connected = false;
  }

  discoverService() {
    return new Promise((resolve, reject) => {
      this.adapter.getServices(this.device.instanceId, (err, services) => {
        if (err) {
          reject(Error(`Error discovering services: ${err}.`));
          return;
        }
        this.service = services.find(s => s.uuid === TEST_SERVICE_UUID);
        if (this.service === undefined) {
          reject(Error('Could not find TEST service'));
        } else {
          resolve();
        }
      });
    });
  }

  discoverCharacteristics() {
    return new Promise((resolve, reject) => {
      this.adapter.getCharacteristics(this.service.instanceId, (err, chars) => {
        if (err) {
          reject(Error(`Error discovering the service characteristics: ${err}.`));
          return;
        }

        this.dataChar = chars.find(c => c.uuid === TEST_DATA_UUID);

        if (this.dataChar === undefined) {
          reject(Error('Could not find all characteristics'));
        } else {
          resolve();
        }
      });
    });
  }

  discoverDataCCCD() {
    return new Promise((resolve, reject) => {
      this.adapter.getDescriptors(this.dataChar.instanceId, (err, descrs) => {
        if (err) {
          reject(Error(`Error discovering the data characteristic CCCD: ${err}.`));
          return;
        }

        this.dataCCCD = descrs.find(d => d.uuid === BLE_UUID_CCCD);
        if (this.dataCCCD === undefined) {
          reject(Error('Could not find data CCCD'));
        } else {
          resolve();
        }
      });
    });
  }

  setDataCCCD() {
    return new Promise((resolve, reject) => {
      this.adapter.writeDescriptorValue(this.dataCCCD.instanceId, [1, 0], true, (err) => {
        if (err) {
          reject(Error(`Failed to set data characteristic CCCD: ${err}.`));
          return;
        }

        resolve();
      });
    });
  }
}

class TESTAdapter {
  constructor(port) {
    this.adapter = adapterFactory.createAdapter('v3', port, '');
    this.adapter.on('logMessage', (severity, message) => console.log(`${message}`));
  }

  open() {
    return new Promise((resolve, reject) => {
      this.adapter.open({ baudRate: 1000000, logLevel: 'error' }, (err) => {
        if (err) {
          reject(Error(`Error opening adapter: ${err}.`));
        }

        resolve();
      });
    });
  }

  connectBle(mac) {
    return new Promise((resolve, reject) => {
      const options = {
        scanParams: {
          active: false,
          interval: 100,
          window: 50,
          timeout: 10,
        },
        connParams: {
          min_conn_interval: 7.5,
          max_conn_interval: 7.5,
          slave_latency: 0,
          conn_sup_timeout: 4000,
        },
      };

      const onConnect = (device) => {
        this.adapter.removeListener('deviceConnected', onConnect);
        resolve(device);
      };
      this.adapter.on('deviceConnected', onConnect);

      this.adapter.connect(mac, options, (err) => {
        if (err) {
          this.adapter.removeListener('deviceConnected', onConnect);
          reject(Error(`Error connecting to target device: ${err}.`));
        }
      });
    });
  }

  connectDevice(mac) {
    return this.connectBle(mac).then(device => new DeviceBLE(this.adapter, device).setup());
  }
}

const adapter = new TESTAdapter('COM26');

adapter.open().then(() => {
  console.log('opened adapter');
  return adapter.connectDevice('F7:5F:92:34:8D:14');
}).then((deviceBle) => {
});

